/*
 * Copyright 1996, 1997 Computing Research Labs, New Mexico State University
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE COMPUTING RESEARCH LAB OR NEW MEXICO STATE UNIVERSITY BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT
 * OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
 * THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */
#ifndef lint
static char rcsid[] = "$Id: GTest.c,v 1.2 1997/04/16 15:08:05 mleisher Exp $";
#endif

#include <stdlib.h>
#include "GTestP.h"

static void
#ifndef _NO_PROTO
_XmuttGlyphTestCheckClass(Widget w)
#else
_XmuttGlyphTestCheckClass(w)
Widget w;
#endif
{
    if (XtIsSubclass(w, xmuttGlyphTestWidgetClass) == False) {
        fprintf(stderr,
                "XmuttGlyphTest: %s not an XmuttGlyphTestWidget.\n",
                XtName(w));
        exit(1);
    }
}

static Dimension
#ifndef _NO_PROTO
_XmuttGlyphTestPMargins(XmuttGlyphTestWidget gw)
#else
_XmuttGlyphTestPMargins(gw)
XmuttGlyphTestWidget gw;
#endif
{
    return (gw->primitive.shadow_thickness << 2) +
        (gw->primitive.highlight_thickness << 1);
}

/*
 * Change window highlighting when focus is gained or lost.
 */
static void
#ifndef _NO_PROTO
_XmuttGlyphTestHighlightWindow(Widget w)
#else
_XmuttGlyphTestHighlightWindow(w)
Widget w;
#endif
{
    int i;
    XmuttGlyphTestWidget gw;
    XRectangle hi[4];

    if (!XtIsRealized(w))
      return;

    gw = (XmuttGlyphTestWidget) w;

    /*
     * Top of window.
     */
    hi[0].x = hi[0].y = 0;
    hi[0].width = gw->core.width;
    hi[0].height = gw->primitive.highlight_thickness;

    /*
     * Right side.
     */
    hi[1].x = gw->core.width - gw->primitive.highlight_thickness;
    hi[1].y = hi[0].y + gw->primitive.highlight_thickness;
    hi[1].width = gw->primitive.highlight_thickness;
    hi[1].height = gw->core.height - (gw->primitive.highlight_thickness << 1);

    /*
     * Bottom of window.
     */
    hi[2].x = hi[0].x;
    hi[2].y = gw->core.height - gw->primitive.highlight_thickness;
    hi[2].width = hi[0].width;
    hi[2].height = hi[0].height;

    /*
     * Left side.
     */
    hi[3].x = hi[0].x;
    hi[3].y = hi[1].y;
    hi[3].width = hi[1].width;
    hi[3].height = hi[1].height;

    if (gw->gtest.hasfocus == True)
      XFillRectangles(XtDisplay(w), XtWindow(w), gw->primitive.highlight_GC,
                      hi, 4);
    else {
        for (i = 0; i < 4; i++)
          XClearArea(XtDisplay(w), XtWindow(w), hi[i].x, hi[i].y,
                     hi[i].width, hi[i].height, False);
    }
}

static void
#ifndef _NO_PROTO
_XmuttGlyphTestShadowWindow(Widget w)
#else
_XmuttGlyphTestShadowWindow(w)
Widget w;
#endif
{
    XmuttGlyphTestWidget gw;
    int i;
    XPoint br[3], tl[3];

    gw = (XmuttGlyphTestWidget) w;

    if (!XtIsRealized(w) || gw->primitive.shadow_thickness == 0)
      return;

    /*
     * Lower left corner outside line.
     */
    tl[0].x = gw->primitive.highlight_thickness;
    tl[0].y = gw->core.height - gw->primitive.highlight_thickness - 1;

    /*
     * Upper left corner outside line.
     */
    tl[1].x = tl[0].x;
    tl[1].y = gw->primitive.highlight_thickness;

    /*
     * Upper right corner outside line.
     */
    tl[2].x = gw->core.width - gw->primitive.highlight_thickness - 1;
    tl[2].y = tl[1].y;

    /*
     * Upper right corner outside line.
     */
    br[0].x = tl[2].x;
    br[0].y = tl[2].y + 1;

    /*
     * Lower right corner outside line.
     */
    br[1].x = br[0].x;
    br[1].y = gw->core.height - gw->primitive.highlight_thickness - 1;

    /*
     * Lower left corner outside line.
     */
    br[2].x = gw->primitive.highlight_thickness + 1;
    br[2].y = br[1].y;

    XDrawLines(XtDisplay(w), XtWindow(w), gw->primitive.top_shadow_GC,
               tl, 3, CoordModeOrigin);
    XDrawLines(XtDisplay(w), XtWindow(w), gw->primitive.bottom_shadow_GC,
               br, 3, CoordModeOrigin);

    /*
     * Draw the remaining shadows successively inward.
     */
    for (i = 1; i < gw->primitive.shadow_thickness; i++) {
        tl[0].x++;
        tl[0].y--;
        tl[1].x++;
        tl[1].y++;
        tl[2].x--;
        tl[2].y++;

        br[0].x--;
        br[0].y++;
        br[1].x--;
        br[1].y--;
        br[2].x++;
        br[2].y--;
        XDrawLines(XtDisplay(w), XtWindow(w), gw->primitive.top_shadow_GC,
                   tl, 3, CoordModeOrigin);
        XDrawLines(XtDisplay(w), XtWindow(w), gw->primitive.bottom_shadow_GC,
                   br, 3, CoordModeOrigin);
    }
}

static void
#ifndef _NO_PROTO
_XmuttGlyphTestSetLineSize(XmuttGlyphTestWidget gw)
#else
_XmuttGlyphTestSetLineSize(gw)
XmuttGlyphTestWidget gw;
#endif
{
    unsigned long i;
    Dimension wd, ht;
    bdf_bbx_t *bbx;
    _XmuttGTestLine *lp;

    lp = &gw->gtest.line;

    if ((bbx = gw->gtest.fontbbx) == 0) {
        /*
         * Default the line height to 15 and width to 100.
         */
        ht = 15;
        wd = gw->gtest.mglyphs * 10;
    } else {
        /*
         * Set the line size from the font metrics.
         */
        ht = bbx->ascent + bbx->descent;
        wd = gw->gtest.mglyphs * bbx->width;
    }

    lp->height = ht;
    lp->width = wd;

    /*
     * Reset the current point to the beginning and at the baseline.
     */
    ht = (gw->gtest.fontbbx == 0) ? lp->height : gw->gtest.fontbbx->ascent;

    lp->cpoint.x = (_XmuttGlyphTestPMargins(gw) >> 1) + 2;
    lp->cpoint.y = (_XmuttGlyphTestPMargins(gw) >> 1) + 2 + ht;

    /*
     * Run through the existing glyphs and adjust the horizontal part of the
     * current point.
     */
    for (i = 0; i < lp->glyphs_used; i++) {
        lp->cpoint.x += lp->glyphs[i]->bbx.x_offset;
        lp->cpoint.x += (gw->gtest.font->spacing != BDF_PROPORTIONAL) ?
            gw->gtest.font->bbx.width : lp->glyphs[i]->dwidth;
    }
}

static void
#ifndef _NO_PROTO
_XmuttGlyphTestPreferredGeometry(XmuttGlyphTestWidget gw, Dimension *width,
                                 Dimension *height)
#else
_XmuttGlyphTestPreferredGeometry(gw, width, height)
XmuttGlyphTestWidget gw;
Dimension *width, *height;
#endif
{
    Dimension wd, ht;

    wd = _XmuttGlyphTestPMargins(gw) + 4 + gw->gtest.line.width;
    ht = _XmuttGlyphTestPMargins(gw) + 4 + gw->gtest.line.height;
    wd = MAX(gw->core.width, wd);
    ht = MAX(gw->core.height, ht);
    *width = wd;
    *height = ht;
}

static void
#ifndef _NO_PROTO
_XmuttGlyphTestMakeGCs(Widget w, Boolean force)
#else
_XmuttGlyphTestMakeGCs(w, force)
Widget w;
Boolean force;
#endif
{
    XmuttGlyphTestWidget gw;
    XtGCMask gcm;
    XGCValues gcv;

    gw = (XmuttGlyphTestWidget) w;

    gcm = GCForeground|GCBackground|GCFunction;

    if (gw->gtest.glyphgc == 0 || force == True) {
        if (gw->gtest.glyphgc != 0)
          XFreeGC(XtDisplay(w), gw->gtest.glyphgc);
        gcv.function = GXcopy;
        gcv.foreground = gw->primitive.foreground;
        gcv.background = gw->core.background_pixel;
        gw->gtest.glyphgc = XCreateGC(XtDisplay(w), XtWindow(w), gcm, &gcv);
    }
}

/*
 * Generate a list of points that represents the image bitmap relative
 * to the baseline already set.
 */
static void
#ifndef _NO_PROTO
_XmuttGlyphTestGlyphImage(XmuttGlyphTestWidget gw, bdf_glyph_t *glyph,
                          Position x, Position y)
#else
_XmuttGlyphTestGlyphImage(gw, glyph)
XmuttGlyphTestWidget gw;
bdf_glyph_t *glyph;
#endif
{
    int i, j, bpr, points;
    unsigned char *bmap;
    XPoint *pp;

    gw->gtest.image_used = 0;
    bmap = glyph->bitmap;
    bpr = (glyph->bbx.width + 7) >> 3;

    /*
     * Resize the point list to be the area of the glyph.
     */
    points = glyph->bbx.width * glyph->bbx.height;
    if (points > gw->gtest.image_size) {
        if (gw->gtest.image_size == 0)
          gw->gtest.image = (XPoint *) XtMalloc(sizeof(XPoint) * points);
        else
          gw->gtest.image = (XPoint *)
              XtRealloc((char *) gw->gtest.image, sizeof(XPoint) * points);
        gw->gtest.image_size = points;
    }

    pp = gw->gtest.image;
    for (i = 0; i < glyph->bbx.height; i++) {
        for (j = 0; j < glyph->bbx.width; j++) {
            if (bmap[(i * bpr) + (j >> 3)] & (0x80 >> (j & 7))) {
                pp->x = x + glyph->bbx.x_offset + j;
                pp->y = (y - glyph->bbx.ascent) + i;
                pp++;
            }
        }
    }
    gw->gtest.image_used = pp - gw->gtest.image;
}

static void
#ifndef _NO_PROTO
_XmuttGlyphTestDrawGlyph(Widget w, bdf_glyph_t *glyph, Boolean unencoded)
#else
_XmuttGlyphTestDrawGlyph(w, glyph, unencoded)
Widget w;
bdf_glyph_t *glyph;
Boolean unencoded;
#endif
{
    XmuttGlyphTestWidget gw;

    if (!XtIsRealized(w))
      return;

    gw = (XmuttGlyphTestWidget) w;

    _XmuttGlyphTestGlyphImage(gw, glyph, gw->gtest.line.cpoint.x,
                              gw->gtest.line.cpoint.y);

    XDrawPoints(XtDisplay(w), XtWindow(w), gw->gtest.glyphgc,
                gw->gtest.image, gw->gtest.image_used, CoordModeOrigin);
}

static void
#ifndef _NO_PROTO
_XmuttGlyphTestRedrawGlyphs(Widget w)
#else
_XmuttGlyphTestRedrawGlyphs(w)
Widget w;
#endif
{
    XmuttGlyphTestWidget gw;
    unsigned long i;
    Position x, y;
    _XmuttGTestLine *lp;

    if (!XtIsRealized(w))
      return;

    gw = (XmuttGlyphTestWidget) w;

    x = (_XmuttGlyphTestPMargins(gw) >> 1) + 2;
    y = gw->gtest.line.cpoint.y;
    lp = &gw->gtest.line;
    for (i = 0; i < lp->glyphs_used; i++) {
        _XmuttGlyphTestGlyphImage(gw, lp->glyphs[i], x, y);
        XDrawPoints(XtDisplay(w), XtWindow(w), gw->gtest.glyphgc,
                    gw->gtest.image, gw->gtest.image_used, CoordModeOrigin);
        x += lp->glyphs[i]->bbx.x_offset;
        x += (gw->gtest.font->spacing != BDF_PROPORTIONAL) ?
            gw->gtest.font->bbx.width : lp->glyphs[i]->dwidth;
    }
}

static void
#ifndef _NO_PROTO
_XmuttGlyphTestRedraw(Widget w)
#else
_XmuttGlyphTestRedraw(w)
Widget w;
#endif
{
    XmuttGlyphTestWidget gw;
    XPoint s, e;
    XRectangle clear;

    if (!XtIsRealized(w))
      return;

    gw = (XmuttGlyphTestWidget) w;

    /*
     * Erase the window.
     */
    clear.x = clear.y = (_XmuttGlyphTestPMargins(gw) >> 1) + 2;
    clear.width = gw->core.width - (clear.x << 1);
    clear.height = gw->core.height - (clear.y << 1);
    XClearArea(XtDisplay(w), XtWindow(w), clear.x, clear.y,
               clear.width, clear.height, False);

    /*
     * Redraw the glyphs.
     */
    _XmuttGlyphTestRedrawGlyphs(w);

    /*
     * Draw the baseline if indicated.
     */
    if (gw->gtest.show_baseline == True) {
        s.x = (_XmuttGlyphTestPMargins(gw) >> 1) + 2;
        e.x = gw->core.width - s.x;
        s.y = e.y = gw->gtest.line.cpoint.y;

        XDrawLine(XtDisplay(w), XtWindow(w), gw->gtest.glyphgc, s.x, s.y,
                  e.x, e.y);
    }
}

/**************************************************************************
 *
 * Class methods.
 *
 **************************************************************************/

static void
#ifndef _NO_PROTO
ClassInitialize(void)
#else
ClassInitialize()
#endif
{
}

static void
#ifndef _NO_PROTO
ClassPartInitialize (WidgetClass wc)
#else
ClassPartInitialize (wc)
WidgetClass wc;
#endif
{
}

static void
#ifndef _NO_PROTO
Initialize(Widget req, Widget newone, ArgList args, Cardinal *num_args)
#else
Initialize(req, newone, args, num_args)
Widget req, newone;
ArgList args;
Cardinal *num_args;
#endif
{
    XmuttGlyphTestWidget gw;
    Dimension wd, ht;

    gw = (XmuttGlyphTestWidget) newone;

    /*
     * Initialize the GCs.
     */
    gw->gtest.glyphgc = 0;

    /*
     * Initialize the points list for the glyph images.
     */
    gw->gtest.image_size = gw->gtest.image_used = 0;

    /*
     * Initialize the line structures.
     */
    (void) memset((char *) &gw->gtest.line, 0, sizeof(_XmuttGTestLine));

    gw->gtest.unencoded = False;
    gw->gtest.show_baseline = True;
    gw->gtest.hasfocus = False;

    /*
     * Set the initial geometry.
     */
    _XmuttGlyphTestSetLineSize(gw);
    _XmuttGlyphTestPreferredGeometry(gw, &wd, &ht);
    gw->core.width = wd;
    gw->core.height = ht;
}

static void
#ifndef _NO_PROTO
Resize(Widget w)
#else
Resize(w)
Widget w;
#endif
{
}

static void
#ifndef _NO_PROTO
Realize(Widget w, Mask *valueMask, XSetWindowAttributes *attributes)
#else
Realize(w, valueMask, attributes)
Widget w;
Mask *valueMask;
XSetWindowAttributes *attributes;
#endif
{
    XtCreateWindow(w, (unsigned int) InputOutput, (Visual *) CopyFromParent,
                   *valueMask, attributes);

    _XmuttGlyphTestMakeGCs(w, False);
}

static void
#ifndef _NO_PROTO
Destroy(Widget w)
#else
Destroy(w)
Widget w;
#endif
{
    XmuttGlyphTestWidget gw;

    gw = (XmuttGlyphTestWidget) w;

    if (gw->gtest.fontbbx != 0)
      XtFree((char *) gw->gtest.fontbbx);
    gw->gtest.fontbbx = 0;

    if (gw->gtest.image_size > 0)
      XtFree((char *) gw->gtest.image);
    gw->gtest.image_size = gw->gtest.image_used = 0;

    if (gw->gtest.glyphgc != 0)
      XFreeGC(XtDisplay(w), gw->gtest.glyphgc);
    gw->gtest.glyphgc = 0;

    if (gw->gtest.line.glyphs_size > 0)
      XtFree((char *) gw->gtest.line.glyphs);
    gw->gtest.line.glyphs_size = gw->gtest.line.glyphs_used = 0;
}

static void
#ifndef _NO_PROTO
Redisplay(Widget w, XEvent *event, Region region)
#else
Redisplay(w, event, region)
Widget w;
XEvent *event;
Region region;
#endif
{
    _XmuttGlyphTestHighlightWindow(w);
    _XmuttGlyphTestShadowWindow(w);
    _XmuttGlyphTestRedraw(w);
}

static Boolean
#ifndef _NO_PROTO
SetValues(Widget curr, Widget req, Widget newone,
          ArgList args, Cardinal *num_args)
#else
SetValues(curr, req, newone, args, num_args)
Widget curr, req, newone;
ArgList args;
Cardinal *num_args;
#endif
{
    XmuttGlyphTestWidget ogw, ngw;
    Dimension wd, ht;
    bdf_bbx_t *bbx;
    Boolean redo;

    ogw = (XmuttGlyphTestWidget) curr;
    ngw = (XmuttGlyphTestWidget) newone;
    redo = False;

    if (ngw->gtest.show_baseline != ogw->gtest.show_baseline)
      redo = True;

    if (ngw->gtest.font != ogw->gtest.font) {
        ngw->gtest.line.glyphs_used = 0;
        (void) memset((char *) &ngw->gtest.line.cpoint, 0, sizeof(XPoint));
        redo = True;
    }

    if (ngw->gtest.fontbbx != ogw->gtest.fontbbx) {
        XtFree((char *) ogw->gtest.fontbbx);
        bbx = ngw->gtest.fontbbx;
        ngw->gtest.fontbbx = (bdf_bbx_t *) XtMalloc(sizeof(bdf_bbx_t));
        (void) memcpy((char *) ngw->gtest.fontbbx, (char *) bbx,
                      sizeof(bdf_bbx_t));

        _XmuttGlyphTestSetLineSize(ngw);
        _XmuttGlyphTestPreferredGeometry(ngw, &wd, &ht);

        wd = MAX(wd, ngw->core.width);
        ht = MAX(ht, ngw->core.height);
        ngw->core.width = wd;
        ngw->core.height = ht;

        redo = True;
    }

    return redo;
}

/**************************************************************************
 *
 * API.
 *
 **************************************************************************/

void
#ifndef _NO_PROTO
XmuttGlyphTestAddGlyph(Widget w, bdf_glyph_t *glyph, Boolean unencoded)
#else
XmuttGlyphTestAddGlyph(w, glyph, unencoded)
Widget w;
bdf_glyph_t *glyph;
Boolean unencoded;
#endif
{
    XmuttGlyphTestWidget gw;
    _XmuttGTestLine *lp;
    Position end, dwidth;

    _XmuttGlyphTestCheckClass(w);

    if (glyph == 0)
      return;

    gw = (XmuttGlyphTestWidget) w;

    /*
     * If the glyph will cause the current point to exceed the edge, ring the
     * bell and return.
     */
    end = gw->core.width - ((_XmuttGlyphTestPMargins(gw) >> 1) + 2);
    dwidth = (gw->gtest.font->spacing != BDF_PROPORTIONAL) ?
        gw->gtest.font->bbx.width : glyph->dwidth;
    if (gw->gtest.line.cpoint.x + glyph->bbx.x_offset + dwidth > end) {
        XBell(XtDisplay(w), 0);
        return;
    }

    /*
     * If the widget is realized, draw the glyph at the current point.
     */
    _XmuttGlyphTestDrawGlyph(w, glyph, unencoded);

    /*
     * Add the glyph to the list and advance the current point.
     */
    lp = &gw->gtest.line;
    if (lp->glyphs_used == lp->glyphs_size) {
        if (lp->glyphs_size == 0)
          lp->glyphs = (bdf_glyph_t **) XtMalloc(sizeof(bdf_glyph_t *) << 3);
        else
          lp->glyphs = (bdf_glyph_t **)
              XtRealloc((char *) lp->glyphs,
                        sizeof(bdf_glyph_t *) * (lp->glyphs_size + 8));
        lp->glyphs_size += 8;
    }
    lp->glyphs[lp->glyphs_used++] = glyph;
    lp->cpoint.x += glyph->bbx.x_offset;
    lp->cpoint.x += (gw->gtest.font->spacing != BDF_PROPORTIONAL) ?
        gw->gtest.font->bbx.width : glyph->dwidth;

    /*
     * Last, call the add callback.
     */
    XtCallCallbackList(w, gw->gtest.add, 0);
}

void
#ifndef _NO_PROTO
XmuttGlyphTestErase(Widget w)
#else
XmuttGlyphTestErase(w)
Widget w;
#endif
{
    XmuttGlyphTestWidget gw;

    _XmuttGlyphTestCheckClass(w);

    gw = (XmuttGlyphTestWidget) w;

    gw->gtest.line.glyphs_used = 0;
    gw->gtest.line.cpoint.x = (_XmuttGlyphTestPMargins(gw) >> 1) + 2;

    _XmuttGlyphTestRedraw(w);
}

void
#ifndef _NO_PROTO
XmuttGlyphTestRedisplay(Widget w)
#else
XmuttGlyphTestRedisplay(w)
Widget w;
#endif
{
    _XmuttGlyphTestCheckClass(w);

    _XmuttGlyphTestRedraw(w);
}

/**************************************************************************
 *
 * Action functions.
 *
 **************************************************************************/

static void
#ifndef _NO_PROTO
GainFocus(Widget w, XEvent *event, String *params, Cardinal *num_params)
#else
GainFocus(w, event, params, num_params)
Widget w;
XEvent *event;
String *params;
Cardinal *num_params;
#endif
{
    XmuttGlyphTestWidget gw;

    /*
     * There is only one way to really gain the focus and that is
     * from a send_event.
     */
    if (!event->xfocus.send_event)
      return;

    gw = (XmuttGlyphTestWidget) w;
    gw->gtest.hasfocus = True;

    _XmuttGlyphTestHighlightWindow(w);
}

static void
#ifndef _NO_PROTO
LoseFocus(Widget w, XEvent *event, String *params, Cardinal *num_params)
#else
LoseFocus(w, event, params, num_params)
Widget w;
XEvent *event;
String *params;
Cardinal *num_params;
#endif
{
    XmuttGlyphTestWidget gw;

    /*
     * There is only one way to really lose the focus and that is
     * from a send_event.
     */
    if (!event->xfocus.send_event)
      return;

    gw = (XmuttGlyphTestWidget) w;
    gw->gtest.hasfocus = False;

    _XmuttGlyphTestHighlightWindow(w);
}

static void
#ifndef _NO_PROTO
Traverse(Widget w, XEvent *event, String *params, Cardinal *num_params)
#else
Traverse(w, event, params, num_params)
Widget w;
XEvent *event;
String *params;
Cardinal *num_params;
#endif
{
    XmuttGlyphTestWidget gw;

    gw = (XmuttGlyphTestWidget) w;

    if (gw->gtest.hasfocus == False)
      return;

    switch (params[0][0]) {
      case 'n':
        (void) XmProcessTraversal(w, XmTRAVERSE_NEXT_TAB_GROUP);
        break;
      case 'p':
        (void) XmProcessTraversal(w, XmTRAVERSE_PREV_TAB_GROUP);
        break;
      case 'r':
        (void) XmProcessTraversal(w, XmTRAVERSE_RIGHT);
        break;
      case 'l':
        (void) XmProcessTraversal(w, XmTRAVERSE_LEFT);
        break;
      case 'u':
        (void) XmProcessTraversal(w, XmTRAVERSE_UP);
        break;
      case 'd':
        (void) XmProcessTraversal(w, XmTRAVERSE_DOWN);
        break;
    }
}

#if defined(XlibSpecificationRelease) && (XlibSpecificationRelease == 6)

static char default_trans[] = "\
<FocusIn>: gain-focus()\n\
<FocusOut>: lose-focus()\n\
~Shift Ctrl ~Meta ~Alt<Key>Right: traverse(right)\n\
~Shift Ctrl ~Meta ~Alt<Key>Left: traverse(left)\n\
~Shift Ctrl ~Meta ~Alt<Key>Up: traverse(up)\n\
~Shift Ctrl ~Meta ~Alt<Key>Down: traverse(down)\n\
~Shift Ctrl ~Meta ~Alt<Key>KP_Right: traverse(right)\n\
~Shift Ctrl ~Meta ~Alt<Key>KP_Left: traverse(left)\n\
~Shift Ctrl ~Meta ~Alt<Key>KP_Up: traverse(up)\n\
~Shift Ctrl ~Meta ~Alt<Key>KP_Down: traverse(down)\n\
Shift ~Ctrl ~Meta ~Alt<Key>Tab: traverse(prev)\n\
~Shift ~Ctrl ~Meta ~Alt<Key>Tab: traverse(next)\n\
";

#else /* R5 or lower */

static char default_trans[] = "\
<FocusIn>: gain-focus()\n\
<FocusOut>: lose-focus()\n\
~Shift Ctrl ~Meta ~Alt<Key>Right: traverse(right)\n\
~Shift Ctrl ~Meta ~Alt<Key>Left: traverse(left)\n\
~Shift Ctrl ~Meta ~Alt<Key>Up: traverse(up)\n\
~Shift Ctrl ~Meta ~Alt<Key>Down: traverse(down)\n\
Shift ~Ctrl ~Meta ~Alt<Key>Tab: traverse(prev)\n\
~Shift ~Ctrl ~Meta ~Alt<Key>Tab: traverse(next)\n\
";

#endif /* R5 or lower */

static XtActionsRec actions[] = {
    {"traverse", Traverse},
    {"gain-focus", GainFocus},
    {"lose-focus", LoseFocus},
};

static XtResource resources[] = {
    {
        XmNtraversalOn,
        XmCTraversalOn,
        XmRBoolean,
        sizeof(Boolean),
        XtOffset(XmPrimitiveWidget, primitive.traversal_on),
        XmRImmediate,
        (XtPointer) True,
    },
    {
        XmNnavigationType,
        XmCNavigationType,
        XmRNavigationType,
        sizeof(unsigned char),
        XtOffset(XmPrimitiveWidget, primitive.navigation_type),
        XmRImmediate,
        (XtPointer) XmTAB_GROUP,
    },
    {
        XmNbdfFont,
        XmCBDFFont,
        XmRPointer,
        sizeof(bdf_font_t *),
        XtOffset(XmuttGlyphTestWidget, gtest.font),
        XmRImmediate,
        0,
    },
    {
        XmNfontBBX,
        XmCFontBBX,
        XmRPointer,
        sizeof(bdf_bbx_t *),
        XtOffset(XmuttGlyphTestWidget, gtest.fontbbx),
        XmRImmediate,
        0,
    },
    {
        XmNglyphsPerLine,
        XmCGlyphsPerLine,
        XmRInt,
        sizeof(int),
        XtOffset(XmuttGlyphTestWidget, gtest.mglyphs),
        XmRImmediate,
        (XtPointer) 10,
    },
    {
        XmNleading,
        XmCLeading,
        XmRInt,
        sizeof(int),
        XtOffset(XmuttGlyphTestWidget, gtest.leading),
        XmRImmediate,
        0,
    },
    {
        XmNglyphDrawDirection,
        XmCGlyphDrawDirection,
        XmRInt,
        sizeof(int),
        XtOffset(XmuttGlyphTestWidget, gtest.dir),
        XmRImmediate,
        (XtPointer) XmuttGLYPHTEST_LTOR,
    },
    {
        XmNshowBaseline,
        XmCShowBaseline,
        XmRBoolean,
        sizeof(Boolean),
        XtOffset(XmuttGlyphTestWidget, gtest.show_baseline),
        XmRImmediate,
        (XtPointer) True,
    },
    {
        XmNaddGlyphCallback,
        XmCAddGlyphCallback,
        XmRCallback,
        sizeof(XtCallbackList),
        XtOffset(XmuttGlyphTestWidget, gtest.add),
        XmRImmediate,
        (XtPointer) 0,
    },
};

externaldef(xmuttglyphtestclassrec)
XmuttGlyphTestClassRec xmuttGlyphTestClassRec = {
    {
	(WidgetClass) &xmPrimitiveClassRec,	/* superclass		*/
	"XmuttGlyphTest",			/* class_name		*/
	sizeof(XmuttGlyphTestRec),		/* widget_size		*/
        ClassInitialize,			/* class_initialize	*/
	ClassPartInitialize,			/* chained class init	*/
	FALSE,					/* class_inited		*/
	Initialize,				/* initialize		*/
        NULL,					/* initialize hook	*/
	Realize,				/* realize		*/
	actions,				/* actions		*/
	XtNumber(actions),			/* num_actions		*/
	resources,				/* resources		*/
	XtNumber(resources),			/* num_resources	*/
	NULLQUARK,				/* xrm_class		*/
	TRUE,					/* compress_motion	*/
	XtExposeCompressMaximal,		/* compress_exposure	*/
        TRUE,					/* compress enter/exit	*/
	FALSE,					/* visible_interest	*/
	Destroy,				/* destroy		*/
	Resize,					/* resize		*/
	Redisplay,				/* expose		*/
	(XtSetValuesFunc) SetValues,		/* set_values		*/
        NULL,					/* set values hook	*/
        XtInheritSetValuesAlmost,		/* set values almost	*/
        NULL,					/* get values hook	*/
	NULL,					/* accept_focus		*/
        XtVersion,				/* version		*/
        NULL,					/* callback offsetlst	*/
        default_trans,				/* default trans	*/
	XtInheritQueryGeometry,			/* query geo proc	*/
	XtInheritDisplayAccelerator,		/* display accelerator	*/
        NULL,					/* extension record reserved*/
    },

    {						/* XmPrimitive		*/
        (XtWidgetProc) _XtInherit,		/* border_highlight	*/
        (XtWidgetProc) _XtInherit,		/* border_unhighlight	*/
        NULL,					/* translations		*/
        NULL,					/* arm_and_activate	*/
        NULL,					/* syn resources	*/
        0,					/* num syn_resources	*/
        0,					/* extension		*/
    },

    {						/* XmuttGlyphTest	*/
        0,					/* Extension record pointer.*/
    }
};

externaldef(xmuttglyphtestwidgetclass)
WidgetClass xmuttGlyphTestWidgetClass = (WidgetClass) &xmuttGlyphTestClassRec;
