/*
 * Copyright 1996, 1997 Computing Research Labs, New Mexico State University
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE COMPUTING RESEARCH LAB OR NEW MEXICO STATE UNIVERSITY BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT
 * OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
 * THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */
#ifndef lint
static char rcsid[] = "$Id: test.c,v 1.2 1997/04/16 15:05:50 mleisher Exp $";
#endif

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <X11/IntrinsicP.h>
#include <Xm/Xm.h>
#include <Xm/RowColumn.h>
#include <Xm/MainW.h>
#include <Xm/CascadeB.h>
#include <Xm/ToggleB.h>
#include <Xm/PushB.h>
#include <Xm/Label.h>
#include <Xm/Separator.h>
#include <Xm/Frame.h>
#include <Xm/Form.h>
#include <Xm/TextF.h>
#include <Xm/List.h>
#include <Xm/DialogS.h>
#include "FGrid.h"
#include "xmbdfed.h"
#include "GTest.h"

static void
#ifndef _NO_PROTO
ClearGlyphTest(Widget w, XtPointer client_data, XtPointer call_data)
#else
ClearGlyphTest(w, client_data, call_data)
Widget w;
XtPointer client_data, call_data;
#endif
{
    MXFEditor *ed;

    ed = &editors[(unsigned long) client_data];
    XmuttGlyphTestErase(ed->test.test);

    XtSetSensitive(ed->test.clear, False);
}

static void
#ifndef _NO_PROTO
ToggleBaseline(Widget w, XtPointer client_data, XtPointer call_data)
#else
ToggleBaseline(w, client_data, call_data)
Widget w;
XtPointer client_data, call_data;
#endif
{
    MXFEditor *ed;
    XmString s;
    Arg av[1];

    ed = &editors[(unsigned long) client_data];

    if (ed->test.show_baseline == True) {
        s = XmStringCreateSimple("Baseline On");
        ed->test.show_baseline = False;
    } else {
        s = XmStringCreateSimple("Baseline Off");
        ed->test.show_baseline = True;
    }

    XtSetArg(av[0], XmNlabelString, s);
    XtSetValues(ed->test.baseline, av, 1);
    XmStringFree(s);

    XtSetArg(av[0], XmNshowBaseline, ed->test.show_baseline);
    XtSetValues(ed->test.test, av, 1);
}

static void
#ifndef _NO_PROTO
EnableClear(Widget w, XtPointer client_data, XtPointer call_data)
#else
EnableClear(w, client_data, call_data)
Widget w;
XtPointer client_data, call_data;
#endif
{
    MXFEditor *ed;

    ed = &editors[(unsigned long) client_data];
    XtSetSensitive(ed->test.clear, True);
}

static void
#ifndef _NO_PROTO
CloseGlyphTest(Widget w, XtPointer client_data, XtPointer call_data)
#else
CloseGlyphTest(w, client_data, call_data)
Widget w;
XtPointer client_data, call_data;
#endif
{
    MXFEditor *ed;

    ed = &editors[(unsigned long) client_data];

    ed->test.active = False;
    XtPopdown(ed->test.shell);
}

static void
#ifndef _NO_PROTO
BuildTestDialog(MXFEditor *ed)
#else
BuildTestDialog(ed)
MXFEditor *ed;
#endif
{
    Widget pb, form, frame, label;
    Cardinal ac;
    Arg av[10];

    if (ed->file == 0)
      sprintf(title, "%s - Test Glyphs: (unnamed%ld)", app_name, ed->id);
    else
      sprintf(title, "%s - Test Glyphs: %s", app_name, ed->file);

    ac = 0;
    XtSetArg(av[ac], XmNtitle, title); ac++;
    XtSetArg(av[ac], XmNallowShellResize, True); ac++;
    XtSetArg(av[ac], XmNdeleteResponse, XmUNMAP); ac++;
    sprintf(name, "xmbdfed_glyphtest%ld", ed->id);
    ed->test.shell = XtCreatePopupShell(name, xmDialogShellWidgetClass, top,
                                        av, ac);

    ac = 0;
    XtSetArg(av[ac], XmNverticalSpacing, 4); ac++;
    XtSetArg(av[ac], XmNmarginWidth, 2); ac++;
    sprintf(name, "xmbdfed_glyphtest%ld_form", ed->id);
    ed->test.form = XtCreateWidget(name, xmFormWidgetClass, ed->test.shell,
                                   av, ac);

    ac = 0;
    XtSetArg(av[ac], XmNtopAttachment, XmATTACH_FORM); ac++;
    XtSetArg(av[ac], XmNleftAttachment, XmATTACH_FORM); ac++;
    XtSetArg(av[ac], XmNrightAttachment, XmATTACH_FORM); ac++;
    sprintf(name, "xmbdfed_glyphtest%ld_frame", ed->id);
    frame = XtCreateManagedWidget(name, xmFrameWidgetClass, ed->test.form,
                                  av, ac);

    ac = 0;
    XtSetArg(av[ac], XmNchildType, XmFRAME_TITLE_CHILD); ac++;
    label = XtCreateManagedWidget("Test Glyphs", xmLabelWidgetClass, frame,
                                  av, ac);

    sprintf(name, "xmbdfed_glyphtest%ld_testform", ed->id);
    form = XtCreateManagedWidget(name, xmFormWidgetClass, frame, 0, 0);

    ac = 0;
    XtSetArg(av[ac], XmNtopAttachment, XmATTACH_FORM); ac++;
    XtSetArg(av[ac], XmNleftAttachment, XmATTACH_FORM); ac++;
    XtSetArg(av[ac], XmNrightAttachment, XmATTACH_FORM); ac++;
    XtSetArg(av[ac], XmNbottomAttachment, XmATTACH_FORM); ac++;
    XtSetArg(av[ac], XmNbdfFont, XmuttFontGridFont(ed->fgrid)); ac++;
    sprintf(name, "xmbdfed_glyphtest%ld_test", ed->id);
    ed->test.test = XtCreateManagedWidget(name, xmuttGlyphTestWidgetClass,
                                          form, av, ac);
    XtAddCallback(ed->test.test, XmNaddGlyphCallback, EnableClear,
                  (XtPointer) ed->id);

    ac = 0;
    XtSetArg(av[ac], XmNtopAttachment, XmATTACH_WIDGET); ac++;
    XtSetArg(av[ac], XmNtopWidget, frame); ac++;
    XtSetArg(av[ac], XmNleftAttachment, XmATTACH_FORM); ac++;
    XtSetArg(av[ac], XmNrightAttachment, XmATTACH_FORM); ac++;
    XtSetArg(av[ac], XmNbottomAttachment, XmATTACH_FORM); ac++;
    sprintf(name, "xmbdfed_glyphtest%ld_buttonframe", ed->id);
    frame = XtCreateManagedWidget(name, xmFrameWidgetClass, ed->test.form,
                                  av, ac);

    ac = 0;
    XtSetArg(av[ac], XmNfractionBase, 3); ac++;
    sprintf(name, "xmbdfed_glyphtest%ld_buttonform", ed->id);
    form = XtCreateManagedWidget(name, xmFormWidgetClass, frame, av, ac);

    ac = 0;
    XtSetArg(av[ac], XmNtopAttachment, XmATTACH_FORM); ac++;
    XtSetArg(av[ac], XmNbottomAttachment, XmATTACH_FORM); ac++;
    XtSetArg(av[ac], XmNleftAttachment, XmATTACH_POSITION); ac++;
    XtSetArg(av[ac], XmNleftPosition, 0); ac++;
    XtSetArg(av[ac], XmNrightAttachment, XmATTACH_POSITION); ac++;
    XtSetArg(av[ac], XmNrightPosition, 1); ac++;
    XtSetArg(av[ac], XmNsensitive, False); ac++;
    ed->test.clear = XtCreateManagedWidget("Clear", xmPushButtonWidgetClass,
                                            form, av, ac);
    XtAddCallback(ed->test.clear, XmNactivateCallback, ClearGlyphTest,
                  (XtPointer) ed->id);

    /*
     * Mark the baseline as being on by default.
     */
    ed->test.show_baseline = True;

    ac = 0;
    XtSetArg(av[ac], XmNtopAttachment, XmATTACH_FORM); ac++;
    XtSetArg(av[ac], XmNbottomAttachment, XmATTACH_FORM); ac++;
    XtSetArg(av[ac], XmNleftAttachment, XmATTACH_POSITION); ac++;
    XtSetArg(av[ac], XmNleftPosition, 1); ac++;
    XtSetArg(av[ac], XmNrightAttachment, XmATTACH_POSITION); ac++;
    XtSetArg(av[ac], XmNrightPosition, 2); ac++;
    ed->test.baseline = XtCreateManagedWidget("Baseline Off",
                                              xmPushButtonWidgetClass,
                                              form, av, ac);
    XtAddCallback(ed->test.baseline, XmNactivateCallback, ToggleBaseline,
                  (XtPointer) ed->id);

    ac = 0;
    XtSetArg(av[ac], XmNtopAttachment, XmATTACH_FORM); ac++;
    XtSetArg(av[ac], XmNbottomAttachment, XmATTACH_FORM); ac++;
    XtSetArg(av[ac], XmNleftAttachment, XmATTACH_POSITION); ac++;
    XtSetArg(av[ac], XmNleftPosition, 2); ac++;
    XtSetArg(av[ac], XmNrightAttachment, XmATTACH_POSITION); ac++;
    XtSetArg(av[ac], XmNrightPosition, 3); ac++;
    pb = XtCreateManagedWidget("Close", xmPushButtonWidgetClass, form, av, ac);
    XtAddCallback(pb, XmNactivateCallback, CloseGlyphTest, (XtPointer) ed->id);
}

void
#ifndef _NO_PROTO
DoGlyphTest(Widget w, XtPointer client_data, XtPointer call_data)
#else
DoGlyphTest(w, client_data, call_data)
Widget w;
XtPointer client_data, call_data;
#endif
{
    MXFEditor *ed;
    bdf_font_t *font;
    Arg av[1];

    ed = &editors[(unsigned long) client_data];

    /*
     * If the glyph test dialog is already up, pop it down.
     */
    if (ed->test.active == True) {
        ed->test.active = False;
        XtPopdown(ed->test.shell);
        return;
    }

    if (ed->test.shell == 0)
      BuildTestDialog(ed);

    /*
     * Reset the dialog title.
     */
    if (ed->file == 0)
      sprintf(title, "%s - Test Glyphs: (unnamed%ld)", app_name, ed->id);
    else
      sprintf(title, "%s - Test Glyphs: %s", app_name, ed->file);

    XtSetArg(av[0], XmNtitle, title);
    XtSetValues(ed->test.shell, av, 1);

    /*
     * Get the font from the font grid and set it in the test widget.
     */
    font = XmuttFontGridFont(ed->fgrid);
    XtSetArg(av[0], XmNfontBBX, &font->bbx);
    XtSetValues(ed->test.test, av, 1);

    /*
     * Indicate that the glyph test dialog is active.  This is used elsewhere
     * to keep the font grid in synch with the glyph tester.
     */
    ed->test.active = True;

    XtManageChild(ed->test.form);
    XtPopup(ed->test.shell, XtGrabNone);
}
